"""energy sum catch div/0

Revision ID: 346125ed062a
Revises: 10f1d9517df7
Create Date: 2020-10-29 17:54:38.794461

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "346125ed062a"
down_revision = "10f1d9517df7"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
    create or replace function interval_size(bucket_interval text, sample_count numeric)
    returns numeric
    immutable
    language plpgsql
    as $$
    begin
        if sample_count > 0 then
            return extract(epoch FROM bucket_interval::interval) / 60 / sample_count;
        end if;

        return 0;
    end;
    $$;

    """
    )


def downgrade():
    pass



"""Small sample error in interval size

Revision ID: e68145e0b0e3
Revises: 346125ed062a
Create Date: 2020-10-31 02:03:05.421703

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "e68145e0b0e3"
down_revision = "346125ed062a"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
    create or replace function interval_size(bucket_interval text, sample_count numeric)
    returns numeric
    immutable
    language plpgsql
    as $$
    begin
        if sample_count > 0 then
            return extract(epoch FROM bucket_interval::interval) / 60 / (sample_count - 1);
        end if;

        return 0;
    end;
    $$;

    """
    )


def downgrade():
    pass



from datetime import datetime
from enum import Enum
from typing import List, Optional

from pydantic import Field

from opennem.api.schema import ApiBase
from opennem.core.dispatch_type import DispatchType
from opennem.schema.opennem import (
    FacilityStatusSchema,
    FueltechSchema,
    NetworkSchema,
)


class FacilityRecord(ApiBase):
    id: int

    network: NetworkSchema

    fueltech: Optional[FueltechSchema]

    status: Optional[FacilityStatusSchema]

    station_id: Optional[int]

    # @TODO no longer optional
    code: str = ""

    dispatch_type: DispatchType
    capacity_registered: Optional[float]

    registered: Optional[datetime]
    deregistered: Optional[datetime]

    network_region: Optional[str]

    unit_id: Optional[int]
    unit_number: Optional[int]
    unit_alias: Optional[str]
    unit_capacity: Optional[float]

    created_by: Optional[str]
    created_at: Optional[datetime]

    approved: bool = False
    approved_by: Optional[str]
    approved_at: Optional[datetime]


class FacilityUpdateResponse(ApiBase):
    success: bool = False
    record: FacilityRecord


class FacilityResponse(ApiBase):
    success: bool = False
    record_num: int
    records: Optional[List[FacilityRecord]]


class FacilityModificationTypes(str, Enum):
    approve = "approve"
    reject = "reject"


class FacilityModification(ApiBase):
    comment: Optional[str] = Field(None)
    modification: FacilityModificationTypes




import logging
import re
from typing import Optional

from pydantic import BaseModel

from opennem.core.unit_single import facility_unit_numbers_are_single

logger = logging.getLogger(__name__)

__is_number = re.compile(r"^\d+$")

__is_single_number = re.compile(r"^\d$")

__is_unit_alias = re.compile(r"([A-Z|a-z]{1,6})")

__is_unit_alias_forced = re.compile(r"([A-Z|a-z]{1,6}\d{1,2})")

__unit_range_parse = re.compile(r"(\d+)\-(\d+)")


class UnitSchema(BaseModel):
    # The unit id
    id: int = 1

    # The unit alias
    alias: Optional[str] = None

    # The number of units
    number: int = 1

    capacity: Optional[int]


is_number = lambda v: bool(re.match(__is_number, v))

is_single_number = lambda v: bool(re.match(__is_single_number, v))

# Has a unit alias like A or GT
unit_has_alias = lambda v: bool(re.search(__is_unit_alias, v))

unit_has_alias_forced = lambda v: bool(re.search(__is_unit_alias_forced, v))

strip_whitespace = lambda v: str(re.sub(r"\s+", "", v.strip()))


def parse_unit_duid(unit_input: str, unit_duid: str):
    return parse_unit_number(
        unit_input, facility_unit_numbers_are_single(unit_duid)
    )


def parse_unit_number(unit_input: str, force_single: bool = False):
    """
        Parses unit number string into a UnitSchema model

        force_single is a hack for units like Hallett where "GT 2-4" means
        unit alias GT2-4 rather than alias GT with id 2 and 2 units

        AEMO put a unit no in sometimes when they mean a unit ID (ie. 8) and
        sometimes it means the number of units (ie. 40)
    """
    unit_id = 1
    unit_no = 0
    unit_alias = None

    has_alias = False

    if unit_input == None:
        unit_input = ""

    # Normalize to string
    if type(unit_input) is not str:
        unit_input = str(unit_input)

    # Strip whitespace and capitalize
    unit_input = strip_whitespace(unit_input)
    unit_input = unit_input.upper()

    if unit_input == "":
        unit_input = "1"

    # @TODO handle the silly multi unit lines
    if "," in unit_input:
        uc = unit_input.split(",")

        # This is a bit of a hack - we use the first unit and
        # count the additionals as extra unit numbers. It works
        # for now
        unit_input = uc[0]
        uc = uc[1:]
        unit_no += len(uc)

        for unit_component in uc:
            if "&" in unit_component:
                unit_no += 1

    if "&" in unit_input:
        unit_no += len(unit_input.split("&"))

    if force_single and unit_has_alias_forced(unit_input):
        has_alias = True

        # extract the unit alias
        unit_alias_search = re.search(__is_unit_alias_forced, unit_input)

        if unit_alias_search and unit_alias_search.lastindex == 1:
            unit_alias = unit_alias_search.group(1)

        if not unit_alias or not type(unit_alias) is str:
            raise Exception(
                "Error extracting alias from {}: Got {}".format(
                    unit_input, unit_alias
                )
            )

        # remove the unit alias
        unit_input = re.sub(r"[A-Za-z]{1,6}\d{1,2}\-", "", unit_input)

    if not has_alias and unit_has_alias(unit_input):
        has_alias = True

        # extract the unit alias
        unit_alias_search = re.search(__is_unit_alias, unit_input)

        if unit_alias_search and unit_alias_search.lastindex == 1:
            unit_alias = unit_alias_search.group(1)

        if not unit_alias or not type(unit_alias) is str:
            raise Exception(
                "Error extracting alias from {}: Got {}".format(
                    unit_input, unit_alias
                )
            )

        # remove the unit alias
        unit_input = re.sub(r"[A-Za-z\ ]", "", unit_input)

    # Simple single number matches
    if is_number(unit_input):
        unit_id = int(unit_input)
        unit_no += 1

        # This is the crazy hack for when AEMO mix unit_no and unit_id
        # in the same field
        if unit_id > 8:
            unit_id = 1
            unit_no = unit_id

    # Range matches (ex. 1-50)
    unit_range_match = re.search(__unit_range_parse, unit_input)

    if unit_range_match and unit_range_match.lastindex == 2:
        unit_id = int(unit_range_match.group(1))
        unit_max = int(unit_range_match.group(2))

        if unit_max < unit_id:
            raise Exception(
                "Invalid max unit number {} on id {} for range {}".format(
                    unit_max, unit_id, unit_input
                )
            )

        unit_no += unit_max - unit_id + 1

    unit = UnitSchema(id=unit_id, number=unit_no, alias=unit_alias,)

    return unit



"""
    support zip files and embedded nested zip files using just an overriden open
    command that supports http endpoints

"""

import io
import logging
import zipfile

from smart_open import open, register_compressor

smart_open_logger = logging.getLogger("smart_open")
smart_open_logger.setLevel(logging.INFO)

urllib_logger = logging.getLogger("urllib3")
urllib_logger.setLevel(logging.INFO)


def chain_streams(streams, buffer_size=io.DEFAULT_BUFFER_SIZE):
    """
    Chain an iterable of streams together into a single buffered stream.
    Usage:
        def generate_open_file_streams():
            for file in filenames:
                yield open(file, 'rb')
        f = chain_streams(generate_open_file_streams())
        f.read()
    """

    class ChainStream(io.RawIOBase):
        def __init__(self):
            self.leftover = b""
            self.stream_iter = iter(streams)
            try:
                self.stream = next(self.stream_iter)
            except StopIteration:
                self.stream = None

        def readable(self):
            return True

        def _read_next_chunk(self, max_length):
            # Return 0 or more bytes from the current stream, first returning all
            # leftover bytes. If the stream is closed returns b''
            if self.leftover:
                return self.leftover
            elif self.stream is not None:
                return self.stream.read(max_length)
            else:
                return b""

        def readinto(self, b):
            buffer_length = len(b)
            chunk = self._read_next_chunk(buffer_length)
            while len(chunk) == 0:
                # move to next stream
                if self.stream is not None:
                    self.stream.close()
                try:
                    self.stream = next(self.stream_iter)
                    chunk = self._read_next_chunk(buffer_length)
                except StopIteration:
                    # No more streams to chain together
                    self.stream = None
                    return 0  # indicate EOF
            output, self.leftover = (
                chunk[:buffer_length],
                chunk[buffer_length:],
            )
            b[: len(output)] = output
            return len(output)

    return io.BufferedReader(ChainStream(), buffer_size=buffer_size)


ZIP_LIMIT = 0


def _handle_zip(file_obj, mode):
    with zipfile.ZipFile(file_obj) as zf:
        if len(zf.namelist()) == 1:
            return zf.open(zf.namelist()[0])

        c = []
        stream_count = 0

        for filename in zf.namelist():
            if filename.endswith(".zip"):
                if ZIP_LIMIT > 0 and stream_count >= ZIP_LIMIT:
                    continue

                c.append(_handle_zip(zf.open(filename), mode))
                stream_count += 1
            else:
                c.append(zf.open(filename))

        return chain_streams(c)


register_compressor(".zip", _handle_zip)



import decimal
import json
from datetime import date, datetime

from geojson import GeoJSONEncoder

from opennem.core.dispatch_type import DispatchType, dispatch_type_string


class OpenNEMJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return float(o)
        if isinstance(o, datetime):
            return o.isoformat()
        if isinstance(o, date):
            return str(o)
        if isinstance(o, DispatchType):
            return dispatch_type_string(o)
        return super(OpenNEMJSONEncoder, self).default(o)


class OpenNEMGeoJSONEncoder(GeoJSONEncoder, OpenNEMJSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return float(o)
        if isinstance(o, datetime):
            return o.isoformat()
        if isinstance(o, DispatchType):
            return dispatch_type_string(o)
        return super(OpenNEMGeoJSONEncoder, self).default(o)


def opennem_deserialize(serialized: str) -> any:

    obj_serialized = None

    # try ujson first because it's faster
    # try:
    #     obj_serialized = ujson.loads(serialized)
    # except TypeError:
    #     pass

    # if not obj_serialized:
    obj_serialized = json.loads(serialized, cls=OpenNEMGeoJSONEncoder)

    return obj_serialized


def opennem_serialize(obj: any, indent=None) -> str:
    obj_deserialized = None

    # try ujson first because it's faster
    # try:
    #     obj_deserialized = ujson.dumps(obj, indent=indent)
    # except TypeError:
    #     pass

    if not obj_deserialized:
        obj_deserialized = json.dumps(
            obj, cls=OpenNEMGeoJSONEncoder, indent=indent
        )

    return obj_deserialized



"""
    Utility functions for API exporter

"""
import json
import logging

from opennem.api.stats.schema import OpennemDataSet
from opennem.exporter.aws import write_to_s3
from opennem.exporter.local import write_to_local
from opennem.settings import settings

logger = logging.getLogger(__name__)


def write_output(
    path: str, stat_set: OpennemDataSet, is_local: bool = False
) -> None:

    if settings.export_local:
        is_local = True

    write_func = write_to_local if is_local else write_to_s3

    if hasattr(stat_set, "json"):
        write_content = stat_set.json(exclude_unset=True)
    else:
        write_content = json.dumps(stat_set)

    byte_count = write_func(path, write_content)

    logger.info("Wrote {} bytes to {}".format(byte_count, path))

    return byte_count



import csv
import json
import os
import sys
import zipfile
from pathlib import Path
from pkgutil import get_data
from typing import Any, List, Optional

MODULE_PATH = Path(os.path.dirname(sys.modules["opennem"].__file__))
DATA_PATH = "core/data"
FIXTURE_PATH = "db/fixtures"
PROJECT_DATA_PATH = "data"

JSON_EXTENSIONS = [".json", ".jsonl", ".geojson"]


class FileNotFound(Exception):
    pass


class FileInvalid(Exception):
    pass


def load_data(
    file_name: str,
    from_project: bool = False,
    from_fixture: bool = False,
    skip_loaders: bool = False,
) -> Any:
    """
        Load a CSV or JSON data file from either the library
        or project data directory

        default loads from `opennem/core/data/`

        from_project is `opennem/data`

        from_fixture is `opennem/db/fixtures`

    """
    data_path = (
        PROJECT_DATA_PATH
        if from_project
        else FIXTURE_PATH
        if from_fixture
        else DATA_PATH
    )

    file_path = Path(file_name)

    data_content = get_data(
        "opennem", os.path.join(data_path, file_name)
    ).decode("utf-8")

    if skip_loaders:
        return data_content

    if file_path.suffix in [".zip"]:
        return load_data_zip(data_content)

    if file_path.suffix in [".csv"]:
        return load_data_csv(data_content)

    if file_path.suffix in JSON_EXTENSIONS:
        return load_data_json(data_content)

    return data_content


def load_data_string(file_path: Path) -> str:

    content = ""

    with open(file_path) as fh:
        content = fh.read()

    return content


def load_data_zip(file_path: Path) -> str:
    content = ""

    with zipfile.ZipFile(file_path) as zf:
        zip_files = zf.namelist()

        if len(zip_files) == 1:
            content = zf.open(zip_files[0]).read()

            if type(content) is bytes:
                content = content.decode("utf-8")

            return {"filename": zip_files[0], "content": content}

        if len(zip_files) != 1:
            raise Exception(
                "Zero or more than one file in zip file. Have {}".format(
                    len(zip_files)
                )
            )


def load_data_json(file_content: str) -> Any:
    fixture: Any = None

    fixture = json.loads(file_content)

    return fixture


def load_data_csv(file_path: Path) -> Optional[List[dict]]:
    """
        Load a CSV file

        @TODO use libmagic to determine encoding types since Excel saves
        can be funky
    """
    records = []

    # leave the encoding in place now todo is to determine it
    with open(file_path, encoding="utf-8-sig") as fh:
        csvreader = csv.DictReader(fh)
        records = [entry for entry in csvreader]

    return records



"""Energy functions for postgres

Revision ID: 7c5a5ddfb9e1
Revises: ad3938cfbd45
Create Date: 2020-10-16 23:10:25.210966

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "7c5a5ddfb9e1"
down_revision = "ad3938cfbd45"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        drop function if exists agg_power (agg_state point, el numeric, interval_size_minutes int);
        create function agg_power(agg_state point, el numeric, interval_size_minutes int)
        returns point
        immutable
        language plpgsql
        as $$
        declare
            current_sum numeric := 0.0;
            num_elements int := 0;
            power_hour float := 0;
            n float := 10;
            w int := 0;
        begin
            num_elements = agg_state[1] + 1;

            IF num_elements = 1 then
                w := 1;
            elsif num_elements = (n + 1) then
                w := 1;
            elsif (num_elements % 2) = 1 then
                w := 2;
            else
                w := 4;
            END IF;

            power_hour := el / (60 / interval_size_minutes);
            current_sum := agg_state[0] + (w * power_hour);

            return point(current_sum, num_elements);
        end;
        $$;
        """
    )

    op.execute(
        """
        drop function if exists onenergy_finalfunc(agg_state point);
        create function onenergy_finalfunc(agg_state point)
        returns float8
        immutable
        strict
        language plpgsql
        as $$
        declare
        result_int numeric = 0.0;
        a float := 0;
        b float := 1;
        n float := 10;
        h float := (b - a) / n;
        begin
            result_int := h * agg_state[0] / 3 * 10;

            return round(result_int, 2);
        end;
        $$;
        """
    )

    op.execute(
        """
        drop aggregate if exists energy_total (numeric, int);
        create aggregate on_energy_sum (numeric, int)
        (
            sfunc = agg_power,
            stype = point,
            finalfunc = onenergy_finalfunc,
            initcond = '(0,0)'
        );
    """
    )


def downgrade():
    op.execute(
        "drop function if exists agg_power (agg_state point, el numeric, interval_size_minutes int)"
    )
    op.execute("drop function if exists onenergy_finalfunc(agg_state point)")
    op.execute("drop aggregate if exists energy_total (numeric, int)")



# pylint: skip-file
import datetime
import re
import time
from io import BytesIO, StringIO

import requests
import sentry_sdk

import pytz
from datadog import statsd
from dms_daemon import orm
from dms_daemon.nemweb_reader import NemFile, ZipFileStreamer

sentry_sdk.init(
    "https://7bdb0e977ab94a1c83db9b927b6dfd29@o402615.ingest.sentry.io/5264004"
)


meta_cols = "DUID, STATION_NAME, PARTICIPANT, REGIONID, FUEL_TECH"
sql = "SELECT {0} FROM nemweb_meta.FULL_REGISTER".format(meta_cols)
cursor = orm.engine_meta.execute(sql)
meta_data = {i[0]: i[1:] for i in cursor.fetchall()}

tz = pytz.timezone("Australia/Melbourne")


class SCADA_File(NemFile):
    # similar to nemfile obj., with additional meta_data funtcionality
    def __init__(
        self,
        fileobject="/data/marble/nemweb/ARCHIVE/DispatchIS_Reports/PUBLIC_DISPATCHIS_20170208.zip",
    ):
        self.meta = fileobject.readline()
        NemFile.__init__(self, fileobject)

    def timestamp(self):
        timestring = ",".join(self.meta.columns[5:7])
        return datetime.datetime.strptime(timestring, "%Y/%m/%d,%H:%M:%S")


def load_scadafile(f="PUBLIC_DISPATCHIS_201309150010_0000000246668929.zip"):
    with ZipFileStreamer(f) as zf:
        if zf.member_count == 1:
            filename = zf.namelist()[0]
            zfs = zf.extract_stream(filename)
            return SCADA_File(zfs)
        else:
            raise Exception("More than one file in zipfile")


def get_page(url="http://www.nemweb.com.au/Reports/CURRENT/Dispatch_SCADA/"):
    response = requests.get(url)
    return response


def get_zip(
    href="/Reports/CURRENT/Dispatch_SCADA/PUBLIC_DISPATCHSCADA_201708161600_0000000286435695.zip",
    log=True,
):
    response = requests.get("http://www.nemweb.com.au{0}".format(href))
    if log:
        web_log(response)
    return BytesIO(response.content)


def web_log(response):
    webtimestamp = response.headers["Last-Modified"]
    timestamp = datetime.datetime.strptime(
        webtimestamp, "%a, %d %b %Y %H:%M:%S GMT"
    ) + datetime.timedelta(0, 10 * 3600)

    Previous_DI = datetime.datetime.now().replace(second=0, microsecond=0)

    # web delay logging
    web_delay = timestamp - Previous_DI
    statsd.gauge("unit.error.web_delay", web_delay.seconds)


def main():
    localtime = datetime.datetime.now(tz).replace(second=0, microsecond=0)
    Previous_DI = localtime - localtime.dst()

    capacity()
    Current_DI = Previous_DI + datetime.timedelta(0, 300)
    DI_timestring = Current_DI.strftime("%Y%m%d%H%M")

    time.sleep(60)
    request_time = datetime.datetime.now()
    page = get_page(
        url="http://www.nemweb.com.au/Reports/CURRENT/Dispatch_SCADA/"
    )
    page2 = get_page(
        url="http://www.nemweb.com.au/Reports/CURRENT/DispatchIS_Reports/"
    )

    pattern = re.compile(
        "/Reports/CURRENT/Dispatch_SCADA/PUBLIC_DISPATCHSCADA_{0}_[0-9]{{16}}.zip".format(
            DI_timestring
        )
    )
    links = re.findall(pattern, page.text)
    links2 = link_getter(page2, DI_timestring)

    if len(links) == 1:
        zipIO = get_zip(href=links[0])
        scada_file = load_scadafile(zipIO)

        zipIO2 = get_zip(href=links2[0])
        dispatch_file = load_scadafile(zipIO2)

        # timedelay logging
        nemde_delay = tz.localize(scada_file.timestamp()) - Previous_DI
        statsd.gauge("unit.error.nemde_delay", nemde_delay.seconds)

        # data processing
        dispatch_scada_proc(scada_file, dispatch_file)
        dispatch_file_proc(dispatch_file)

    elif len(links) == 0:
        statsd.gauge("unit.error.skipped_di", 1)


def dispatch_scada_proc(sf, df):
    for i, row in sf.DISPATCH_UNIT_SCADA.iterrows():
        try:
            meta = meta_data[row.DUID]
            region = meta[-2]
            price = df.DISPATCH_PRICE.RRP[
                df.DISPATCH_PRICE.REGIONID == region
            ].values[0]
            scadavalue = row.SCADAVALUE
            meta_pairs = [
                "{0}:{1}".format(i, j)
                for i, j in zip(meta_cols.lower().split(","), meta)
            ]
            statsd.gauge("unit.power", scadavalue, tags=meta_pairs)
            statsd.gauge(
                "unit.marketvalue", scadavalue * price / 12.0, tags=meta_pairs
            )
        except:
            print("missing")
            pass


def dispatch_file_proc(df):
    dp = df.DISPATCH_PRICE[["REGIONID", "RRP"]]
    dr = df.DISPATCH_REGIONSUM[["REGIONID", "TOTALDEMAND", "NETINTERCHANGE"]]
    dc = dp.merge(dr, on="REGIONID")
    dc["VALUE"] = dc.RRP * dc.TOTALDEMAND / 12.0
    for i, row in dc.iterrows():
        statsd.gauge(
            "dispatch.totaldemand",
            row.TOTALDEMAND,
            tags=["regionid:{0}".format(row.REGIONID)],
        )
        statsd.gauge(
            "dispatch.netinterchange",
            row.NETINTERCHANGE,
            tags=["regionid:{0}".format(row.REGIONID)],
        )
        statsd.gauge(
            "dispatch.price",
            row.RRP,
            tags=["regionid:{0}".format(row.REGIONID)],
        )
        statsd.gauge(
            "dispatch.marketvalue",
            row.VALUE,
            tags=["regionid:{0}".format(row.REGIONID)],
        )


def link_getter(page, DI_timestring):
    pattern = re.compile(
        "/Reports/CURRENT/DispatchIS_Reports/PUBLIC_DISPATCHIS_{0}_[0-9]{{16}}.zip".format(
            DI_timestring
        )
    )
    links = re.findall(pattern, page.text)
    return links


def capacity():
    cap_sql = sql.format("REG_CAP")
    cursor = orm.engine_meta.execute(cap_sql)
    cap_data = {i[0]: i[1:] for i in cursor.fetchall()}

    for duid in cap_data:
        value = cap_data[duid][0]
        meta = meta_data[duid]
        meta_pairs = [
            "{0}:{1}".format(i, j)
            for i, j in zip(meta_cols.lower().split(","), meta)
        ]
        statsd.gauge("unit.capacity", value, meta_pairs)


def main_test():
    Previous_DI = datetime.datetime.now().replace(second=0, microsecond=0)
    capacity()
    Current_DI = Previous_DI + datetime.timedelta(0, 300)
    DI_timestring = Current_DI.strftime("%Y%m%d%H%M")

    time.sleep(60)
    request_time = datetime.datetime.now()

    page = get_page(
        url="http://www.nemweb.com.au/Reports/CURRENT/Dispatch_SCADA/"
    )
    page2 = get_page(
        url="http://www.nemweb.com.au/Reports/CURRENT/DispatchIS_Reports/"
    )

    pattern = re.compile(
        "/Reports/CURRENT/Dispatch_SCADA/PUBLIC_DISPATCHSCADA_{0}_[0-9]{{16}}.zip".format(
            DI_timestring
        )
    )
    links = re.findall(pattern, page.text)
    links2 = link_getter(page2, DI_timestring)

    zipIO = get_zip(href=links[0])
    zipIO2 = get_zip(href=links2[0])

    return load_scadafile(zipIO), load_scadafile(zipIO2)


if __name__ == "__main__":
    main()



"""account for one sample in energy_sum

Revision ID: d8e3e95a80d1
Revises: e68145e0b0e3
Create Date: 2020-10-31 02:09:35.285592

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "d8e3e95a80d1"
down_revision = "e68145e0b0e3"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        create or replace function interval_size(bucket_interval text, sample_count numeric)
        returns numeric
        immutable
        language plpgsql
        as $$
        begin
            if sample_count > 1 then
                return extract(epoch FROM bucket_interval::interval) / 60 / (sample_count - 1);
            end if;

            if sample_count = 1 then
                return extract(epoch FROM bucket_interval::interval) / 60;
            end if;

            return 0;
        end;
        $$;

    """
    )


def downgrade():
    pass



from typing import List

import pytz
from fastapi import APIRouter, Depends, HTTPException, Query
from sqlalchemy.orm import Session
from starlette import status

from opennem.api.stats.controllers import get_scada_range, stats_factory
from opennem.api.stats.schema import DataQueryResult, OpennemDataSet
from opennem.api.time import human_to_interval, human_to_period
from opennem.api.weather.queries import observation_query
from opennem.core.networks import network_from_network_code
from opennem.core.units import get_unit
from opennem.db import get_database_engine, get_database_session
from opennem.db.models.opennem import BomStation
from opennem.utils.timezone import get_fixed_timezone

from .schema import WeatherObservation, WeatherStation

router = APIRouter()


@router.get(
    "/station",
    description="List of weather stations",
    response_model=List[WeatherStation],
    response_model_exclude_unset=True,
)
def station(
    session: Session = Depends(get_database_session),
) -> List[WeatherStation]:
    """
        Get a list of all stations

    """
    stations = session.query(BomStation).all()

    return stations


@router.get(
    "/station/{station_code}",
    description="Weather station record",
    response_model=WeatherStation,
    response_model_exclude_unset=True,
    response_model_exclude=set(["observations"]),
)
def station_record(
    station_code: str = Query(..., description="Station code"),
    session: Session = Depends(get_database_session),
) -> WeatherStation:
    """
        Get a single weather station by code

    """
    station = (
        session.query(BomStation)
        .filter(BomStation.code == station_code)
        .one_or_none()
    )

    if not station:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="No station found"
        )

    return station


@router.get(
    "/station/observation/{station_code}",
    description="Observations from a station",
    response_model=OpennemDataSet,
    response_model_exclude_unset=True,
)
def station_observations_api(
    station_code: str = Query(None, description="Station code"),
    interval: str = Query("15m", description="Interval"),
    period: str = Query("7d", description="Period"),
    station_codes: List[str] = [],
    network_code: str = "NEM",
    timezone: str = None,
    offset: str = None,
    year: int = None,
    engine=Depends(get_database_engine),
) -> OpennemDataSet:
    units = get_unit("temperature")

    if not interval:
        interval = "15m"

    if not period:
        period = "7d"

    network = None

    if network_code:
        network = network_from_network_code(network_code)

    if station_code:
        station_codes = [station_code]

    interval = human_to_interval(interval)
    period = human_to_period(period)

    if timezone:
        timezone = pytz.timezone(timezone)

    if offset:
        timezone = get_fixed_timezone(offset)

    scada_range = None

    if network:
        scada_range = get_scada_range(network=network)

    query = observation_query(
        station_codes=station_codes,
        interval=interval,
        period=period,
        network=network,
        scada_range=scada_range,
        year=year,
    )

    with engine.connect() as c:
        results = list(c.execute(query))

    stats = [
        DataQueryResult(
            interval=i[0], result=i[2], group_by=i[1] if len(i) > 1 else None
        )
        for i in results
    ]

    if len(stats) < 1:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Station stats not found",
        )

    result = stats_factory(
        stats=stats,
        units=units,
        interval=interval,
        period=period,
        network=network,
        code="bom",
        group_field="temperature",
    )

    return result



from enum import Enum
from pathlib import Path
from typing import Optional, Union

import tomlkit
from packaging.version import Version, parse


def _get_project_meta():
    pyproject = Path(__file__).parent / "../../pyproject.toml"

    if not pyproject.is_file():
        return None

    with pyproject.open() as pp_file:
        file_contents = pp_file.read()

    return tomlkit.parse(file_contents)["tool"]["poetry"]


pkg_meta = _get_project_meta()
version = None

if pkg_meta:
    project = str(pkg_meta["name"])
    copyright = "2020, OpenNEM"
    version = str(pkg_meta["version"])

release = version


class VersionPart(Enum):
    MAJOR = "major"
    MINOR = "minor"
    PATCH = "patch"


def get_version(
    version_part: Optional[VersionPart] = None, as_string: bool = True
) -> Union[str, Version]:
    if not version:
        return None

    version_parsed = parse(version)

    if not version_part:
        if as_string:
            return version
        return version_parsed

    if version_part == VersionPart.MAJOR:
        return version_parsed.major

    if version_part == VersionPart.MINOR:
        return version_parsed.minor

    if version_part == VersionPart.PATCH:
        return version_parsed.minor



# GULLRWF2 -> Biala
import json
from datetime import date, datetime
from pprint import pprint

from sqlalchemy.orm import sessionmaker

from opennem.core.unit_codes import get_unit_code
from opennem.core.unit_parser import parse_unit_duid
from opennem.db import db_connect
from opennem.db.models.opennem import Facility, Station, metadata

engine = db_connect()
session = sessionmaker(bind=engine)


def patches():

    sqls = [
        # "update facility set capacity_registered = 2.0, unit_capacity = 2.0 where code = 'GOSNELLS'",
        # "update facility set capacity_registered = 1.1, unit_capacity = 1.1  where code = 'ATLAS'",
        # code GULLRWF2_74 -> Biala
        "update facility set active=false where network_code ='GULLRWF2'",
        "update facility set station_id = (select id from station where name = 'Wivenhoe Small Hydro') where code ='WIVENSH'",
        "update station set name = 'Wivenhoe Mini' where name = 'Wivenhoe Small'",
        # "update facility set fueltech_id = 'pumps' where network_code in ('PUMP2', 'PUMP1')",
        # "update facility set active=false where code='PIONEER'",
        # "update facility set station_id = null where name='Crookwell' and code is null",
        # "update facility set station_id = null where name='Pioneer Sugar Mill' and code is null",
    ]

    with engine.connect() as c:

        for query in sqls:
            rows = c.execute(query)
            pprint(rows)

    s = session()

    duid = None

    unit = parse_unit_duid(1, duid)
    unit_code = get_unit_code(unit, duid, "Singleton Solar Farm")

    singleton = Station(
        name="Singleton",
        locality="singleton",
        network_name="Singleton Solar Farm",
        network_id="NEM",
        created_by="opennem.patches",
    )
    s.add(singleton)

    singleston_facility = Facility(
        code="0NSISF_1",
        status_id="operating",
        network_region="NSW1",
        network_name="Singleton Solar Farm",
        fueltech_id="solar_utility",
        unit_id=unit.id,
        unit_number=unit.number,
        unit_capacity=0.4,
        capacity_registered=0.4,
        created_by="opennem.patches",
    )
    singleston_facility.station = singleton

    s.add(singleston_facility)
    s.commit()


if __name__ == "__main__":
    patches()



import logging
from typing import List

from fastapi import Depends, FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from sqlalchemy.orm import Session
from starlette import status

from opennem.api.admin.router import router as admin_router
from opennem.api.export.router import router as export_router
from opennem.api.facility.router import router as facility_router
from opennem.api.geo.router import router as geo_router
from opennem.api.locations import router as locations_router
from opennem.api.revision.router import router as revisions_router
from opennem.api.station.router import router as station_router
from opennem.api.stats.router import router as stats_router
from opennem.api.weather.router import router as weather_router
from opennem.core.time import INTERVALS, PERIODS
from opennem.core.units import UNITS
from opennem.db import get_database_session
from opennem.db.models.opennem import FuelTech, Network
from opennem.schema.network import NetworkSchema
from opennem.schema.opennem import FueltechSchema
from opennem.schema.time import TimeInterval, TimePeriod
from opennem.schema.units import UnitDefinition
from opennem.settings import settings
from opennem.utils.version import get_version

from .schema import FueltechResponse

logger = logging.getLogger(__name__)

app = FastAPI(
    title="OpenNEM", debug=settings.debug, version=get_version(as_string=True)
)

try:
    from fastapi.staticfiles import StaticFiles

    app.mount(
        "/static",
        StaticFiles(directory=settings.static_folder_path),
        name="static",
    )
except Exception as e:
    logger.error("Error initializing static hosting: {}".format(e))


app.include_router(stats_router, tags=["Stats"], prefix="/stats")
app.include_router(locations_router, tags=["Locations"], prefix="/locations")
app.include_router(geo_router, tags=["Geo"], prefix="/geo")
app.include_router(station_router, tags=["Stations"], prefix="/station")
app.include_router(facility_router, tags=["Facilities"], prefix="/facility")
app.include_router(revisions_router, tags=["Revisions"], prefix="/revision")
app.include_router(weather_router, tags=["Weather"], prefix="/weather")
app.include_router(admin_router, tags=["Admin"], prefix="/admin")
app.include_router(export_router, tags=["Export"], prefix="/export")


origins = [
    "https://opennem.org.au",
    "https://dev.opennem.org.au",
    "https://admin.opennem.org.au",
    "https://admin.opennem.test",
    "http://localhost:8001",
    "http://127.0.0.1:8001",
    "http://127.0.0.1:8002",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/networks", response_model=List[NetworkSchema])
def networks(
    session: Session = Depends(get_database_session),
) -> List[NetworkSchema]:
    networks = session.query(Network).all()

    if not networks:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    response = [NetworkSchema.parse_obj(i) for i in networks]

    return response


@app.get("/fueltechs", response_model=List[FueltechSchema])
def fueltechs(
    session: Session = Depends(get_database_session),
) -> FueltechResponse:
    fueltechs = session.query(FuelTech).all()

    if not fueltechs:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    response = [FueltechSchema.parse_obj(i) for i in fueltechs]

    return response


@app.get("/intervals", response_model=List[TimeInterval])
def intervals() -> List[TimeInterval]:
    return INTERVALS


@app.get("/periods", response_model=List[TimePeriod])
def periods() -> List[TimePeriod]:
    return PERIODS


@app.get("/units", response_model=List[UnitDefinition])
def units() -> List[UnitDefinition]:
    return UNITS



from sqlalchemy.orm import sessionmaker

from opennem.db import db_connect


class DatabaseStoreBase(object):
    def __init__(self):
        self.engine = engine = db_connect()
        self.session = sessionmaker(bind=engine)



import logging

import requests
from pydantic import ValidationError

from opennem.api.export.map import StatMetadata
from opennem.settings import settings
from opennem.utils.url import bucket_to_website, urljoin

logger = logging.getLogger(__name__)


def check_metadata_status() -> bool:
    metadata_path = bucket_to_website(
        urljoin(settings.s3_bucket_path, "metadata.json")
    )

    resp = requests.get(metadata_path)

    if resp.status_code != 200:
        logger.error("Error retrieving opennem metadata")
        return False

    resp_json = resp.json()

    metadata = None

    try:
        metadata = StatMetadata.parse_obj(resp_json)
    except ValidationError as e:
        logger.error("Validation error in metadata: {}".format(e))

    for resource in metadata.resources:
        if not resource.path:
            logger.info("Resource without path")
            continue

        resource_website_path = bucket_to_website(
            urljoin(settings.s3_bucket_path, resource.path)
        )

        r = requests.get(resource_website_path)

        if r.status_code != 200:
            logger.error(
                "Error with metadata resource: {}".format(
                    resource_website_path
                )
            )


if __name__ == "__main__":
    check_metadata_status()



import logging
import os
from io import BytesIO

import requests
from PIL import Image
from smart_open import open

TEST_PHOTO = "https://upload.wikimedia.org/wikipedia/commons/thumb/5/5b/Jervis_Bay_nuclear_plant_site.JPG/650px-Jervis_Bay_nuclear_plant_site.JPG"


S3_EXPORT_DEFAULT_BUCKET = "s3://photos.opennem.org.au/"


UPLOAD_ARGS = {
    "ContentType": "image/jpeg",
}


logging.basicConfig(level=logging.INFO)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

from opennem.db import SessionLocal
from opennem.db.models.opennem import Photo


def write_photo_to_s3(file_path: str, data):
    s3_save_path = os.path.join(S3_EXPORT_DEFAULT_BUCKET, file_path)
    write_count = 0

    with open(
        s3_save_path,
        "wb",
        transport_params=dict(multipart_upload_kwargs=UPLOAD_ARGS),
    ) as fh:
        write_count = fh.write(data)

    logger.info("Wrote {} to {}".format(len(data), s3_save_path))

    return write_count


def photos_process():
    session = SessionLocal()

    photos = session.query(Photo).filter(Photo.processed == False).all()

    # for photo in photos:
    #     r = write_photo_to_s3()


def img_to_buffer(img):
    buf = BytesIO()

    if img.mode in ("RGBA", "P", "LA"):
        img = img.convert("RGB")

    img.save(buf, format="JPEG")

    return buf.getbuffer()


def store_photo(photo_url: str):
    img = Image.open(requests.get(photo_url, stream=True).raw)

    save_bugger = BytesIO()
    img.save(save_bugger, format="JPEG")

    write_photo_to_s3("test.jpeg", data=save_bugger.getbuffer())


if __name__ == "__main__":
    photos_process()



from typing import Optional

from pydantic import Field

from .core import BaseConfig

JOIN_QUERY = """ +
    ((extract(minute FROM fs.{field} AT TIME ZONE '{timezone}')::int / {interval})::integer
    * interval '{interval} minute')::interval
"""


class TimeInterval(BaseConfig):
    # interval size in minutes
    interval: int = Field(..., description="Interval size in minutes")

    interval_human: str = Field(
        ..., description="Interval size in human length"
    )

    interval_sql: str

    trunc: str

    def get_sql_join(
        self, field="trading_interval", timezone="UTC"
    ) -> Optional[str]:
        if self.interval >= 60:
            return ""

        return JOIN_QUERY.format(
            interval=self.interval, field=field, timezone=timezone
        )


class TimeIntervalAPI(BaseConfig):
    interval: int = Field(..., description="Interval size in minutes")
    interval_human: str = Field(
        ..., description="Interval size in human length"
    )


class TimePeriod(BaseConfig):
    period: int = Field(..., description="Period in minutes")

    period_human: str = Field(..., description="Period in human size")

    period_sql: str


class TimePeriodAPI(BaseConfig):
    period: int = Field(..., description="Period in minutes")
    period_human: str = Field(..., description="Period in human size")



import io
import logging
from os import makedirs
from pathlib import Path

from opennem.settings import settings

logger = logging.getLogger(__name__)


def write_to_local(file_path, data):
    save_folder = settings.static_folder_path

    save_file_path = Path(save_folder) / file_path.lstrip("/")

    dir_path = save_file_path.resolve().parent

    if not dir_path.is_dir():
        makedirs(dir_path)

    if type(data) is io.StringIO:
        data = data.getvalue()

    bytes_written = 0

    with open(save_file_path, "w") as fh:
        bytes_written += fh.write(data)

    print("Wrote {} to {}".format(bytes_written, save_file_path))



"""Network regions table

Revision ID: 69449fda6381
Revises: d8e3e95a80d1
Create Date: 2020-11-09 14:42:05.172383

"""
import sqlalchemy as sa
from alembic import op

revision = "69449fda6381"
down_revision = "d8e3e95a80d1"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "network_region",
        sa.Column("created_by", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("network_id", sa.Text(), nullable=False),
        sa.Column("code", sa.Text(), nullable=False),
        sa.Column("timezone", sa.Text(), nullable=True),
        sa.Column("timezone_database", sa.Text(), nullable=True),
        sa.Column("offset", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["network_id"],
            ["network.code"],
            name="fk_network_region_network_code",
        ),
        sa.PrimaryKeyConstraint("network_id", "code"),
    )


def downgrade():
    op.drop_table("network_region")



from functools import wraps
from typing import Dict, List, Optional

from expiringdict import ExpiringDict

from opennem.api.stats.schema import ScadaDateRange
from opennem.schema.network import NetworkSchema

CACHE_AGE = 60 * 15

scada_cache = ExpiringDict(max_len=100, max_age_seconds=CACHE_AGE)


def cache_scada_result(func):
    @wraps(func)
    def _cache_scada_wrapper(
        network: Optional[NetworkSchema] = None,
        networks: Optional[List[NetworkSchema]] = None,
        network_region: Optional[str] = None,
    ):
        key_list = [network.code]

        # if network_region:
        #     if hasattr(network_region, "code"):
        #         key_list.append(network_region.code)
        #     else:
        #         key_list.append(network_region)

        if networks:
            key_list += [n.code for n in networks]

        key = frozenset(key_list)

        try:
            _val: Dict = scada_cache[key]
            ret: ScadaDateRange = ScadaDateRange(**_val)
            return ret
        except KeyError:
            ret: ScadaDateRange = func(network, networks, network_region)
            scada_cache[key] = ret.dict()
            return ret

    return _cache_scada_wrapper
